@@ -19,21 +19,20 @@ from botocore.exceptions import WaiterError
 logging.basicConfig(level=logging.INFO)
 
 
-# exception hook so errors raise to main thread
-def exception_hook(args):
-    traceback.print_exc()
-    globals()['_call_on_all_errors'] = True
-
-
-# set the exception hook
-threading.excepthook = exception_hook
+class TaskThread(threading.Thread):
+    def run(self):
+        try:
+            super(TaskThread, self).run()
+        except Exception as e:
+            traceback.print_exc()
+            os._exit(1)
 
 
 class EC2Manager:
     def __init__(self, **kwargs):
         # needed envs
-        self._repo = os.environ.get('REPO', kwargs.get('repo'))
-        self._github_user = os.environ.get('GITHUB_USERNAME', kwargs.get('github_username'))
+        self.repo = os.environ.get('REPO', kwargs.get('repo'))
+        self.github_user = os.environ.get('GITHUB_USERNAME', kwargs.get('github_username'))
         self._github_token = os.environ.get('GITHUB_TOKEN', kwargs.get('github_token'))
 
         # optional envs
@@ -44,7 +43,7 @@ class EC2Manager:
         )
         self._template_directory = os.path.join(os.path.dirname(__file__), 'template')
         self._repo_url = (
-            f'https://{self._github_user}:{self._github_token}@github.com/{self._github_user}/{self._repo}.git'
+            f'https://{self.github_user}:{self._github_token}@github.com/{self.github_user}/{self.repo}.git'
         )
         self._max_attempts = int(os.environ.get('MAX_TIMEOUT', 600))
 
@@ -56,7 +55,12 @@ class EC2Manager:
         # only when initialize with a config
         if self._config_file:
             # internal variables
-            self._config = self._get_config()
+            self._config = self.get_config()
+
+            # pull values from config if they are not already set
+            self.github_user = self.github_user or self._config.get('github_username')
+            self.repo = self.repo or self._config.get('github_repo')
+
             self._aws_default_region = self._config.get('aws_region')
             self._type = self._config.get('type')
             self._vpc_name = self._config.get('vpc_name')
@@ -89,7 +93,7 @@ class EC2Manager:
             self._aws_backend_bucket = os.environ.get('AWS_BACKEND_BUCKET', f'{account_number}-terraform-state')
             self._aws_backend_key = f'{self._type}.tfstate'
 
-    def _get_config(self):
+    def get_config(self):
         """
         Gets the config data.
 
@@ -102,18 +106,22 @@ class EC2Manager:
         return {}
 
     @staticmethod
-    def _get_encoded_envs(encoded_string):
+    def _get_encoded_envs(encoded_env_data):
         """
         Gets environment variables from the given encoded string.
 
-        :param str encoded_string: A env file encoded as a base64 string.
+        :param dict encoded_env_data: The encoded env file data.
         :returns: A dictionary of environment variables.
         :rtype: dict
         """
+        encoded_string = os.environ.get(encoded_env_data.get('env_var', ''), '')
+
         # decode the env file
         env_file_contents = base64.b64decode(encoded_string).decode('utf-8')
-        return {line.split('=')[0]: line.split('=')[1].strip("'").strip('"') for line in env_file_contents.split('\n')
-                if '=' in line}
+        return {
+            line.split('=')[0]: line.split('=')[1].strip("'").strip('"') for line in env_file_contents.split('\n')
+            if '=' in line and not line.strip().startswith('#')
+        }
 
     @staticmethod
     def encode_string(string):
@@ -129,7 +137,7 @@ class EC2Manager:
         :returns: Working directory path.
         :rtype: str
         """
-        path = os.path.normpath(os.path.join(self._repo, working_directory)).replace(os.path.sep, '/')
+        path = os.path.normpath(os.path.join(self.repo, working_directory)).replace(os.path.sep, '/')
         return f'./{path}'
 
     def _get_instance_id(self, name, try_once=False, attempts=0):
@@ -200,10 +208,11 @@ class EC2Manager:
             'ports': instances[name].get('ports', []) + default_ports,
             'envs': {
                 'NAME': name,
-                'REPO': self._repo,
+                'REPO': self.repo,
                 'COMMIT': self._commit,
+                'GITHUB_USERNAME': self.github_user,
                 **instances[name].get('envs', {}),
-                **self._get_encoded_envs(os.environ.get(instances[name].get('encoded_env_file_variable', ''), ''))
+                **self._get_encoded_envs(instances[name].get('encoded_env_file', {}))
             }
         } for name in self._instance_repo_updates}
 
@@ -293,6 +302,39 @@ class EC2Manager:
             if process.returncode != 0:
                 raise RuntimeError('Failed to apply terraform')
 
+    def _destroy_terraform(self):
+        """
+        Runs terraform destroy.
+        """
+        self._logger.info(f'Destroying resources...')
+        instances = self.encode_string(json.dumps(self._get_instance_data()))
+
+        process = subprocess.run(
+            [
+                'terraform',
+                'destroy',
+                '-auto-approve',
+                '-var',
+                f'type={self._type}',
+                '-var',
+                f'vpc_name={self._vpc_name}',
+                '-var',
+                f'public_subnet_cidr={self._public_subnet_cidr}',
+                '-var',
+                f'private_subnet_cidr={self._private_subnet_cidr}',
+                '-var',
+                f'aws_region={self._aws_default_region}',
+                '-var',
+                f'expose_docker_daemons={self._expose_docker_daemons}',
+                '-var',
+                f'instances={instances}'
+            ],
+            env=os.environ,
+            cwd=self._terraform_directory
+        )
+        if process.returncode != 0:
+            raise RuntimeError('Failed to run terraform destroy')
+
     def _call_on_all(self, callable_instance, items, extra_args=None, **kwargs):
         """
         Helper method for calling a given method on all provided.
@@ -303,8 +345,6 @@ class EC2Manager:
         :param bool join: Whether or not to join the threads.
         :return list[Thread]: A list of threads.
         """
-        globals()['_call_on_all_errors'] = False
-
         if not extra_args:
             extra_args = []
 
@@ -323,7 +363,7 @@ class EC2Manager:
             if data:
                 args.insert(1, data)
 
-            thread = threading.Thread(
+            thread = TaskThread(
                 name=name,
                 target=callable_instance,
                 args=args + extra_args,
@@ -338,10 +378,6 @@ class EC2Manager:
                 thread.join()
             return []
 
-        # stop the main process if a thread created an error
-        if globals()['_call_on_all_errors']:
-            sys.exit(1)
-
         return threads
 
     def _await_instances(self):
@@ -362,18 +398,20 @@ class EC2Manager:
         :rtype: dict
         """
         self._logger.info(f'Exporting Terraform outputs...')
-        output = subprocess.check_output(
-            ['terraform', 'output', '-json'],
-            env=os.environ,
-            cwd=self._terraform_directory
-        )
-        return json.loads(output)
+        try:
+            return json.loads(subprocess.check_output(
+                ['terraform', 'output', '-json'],
+                env=os.environ,
+                cwd=self._terraform_directory
+            ))
+        except:
+            return {}
 
     def get_repo(self):
         """
         Gets the repo instance.
         """
-        return self._github_client.get_repo(full_name_or_id=f'{self._github_user}/{self._repo}')
+        return self._github_client.get_repo(full_name_or_id=f'{self.github_user}/{self.repo}')
 
     def get_docker_logs(self, name, data):
         """
@@ -473,21 +511,37 @@ class EC2Manager:
         # wait for each instance status to be ok
         self._await_instances()
 
+    def _set_env_file(self, name, data):
+        """
+        Writes a env file to disk in the working directory.
+
+        :param str name: The name of the instance.
+        :param dict data: The instance's data.
+        """
+        file_name = self._config['instances'][name].get('encoded_env_file', {}).get('file_name')
+        if file_name:
+            self._logger.info(f'{name} setting env file {file_name}...')
+
+            env_file_content = '\n'.join([f'{key}={value}' for key, value in data['envs'].items()])
+            working_directory = self._get_working_directory(data['working_directory'])
+            self.run_command(name, [
+                f'cd {working_directory}',
+                f'echo "{env_file_content}" > {file_name}'
+            ], print_output=False)
+
     def _update_repo(self, name):
         """
         Updates a single instance repo.
         """
-        if self._repo not in self.run_command(name, ['ls'], print_output=False):
-            self._logger.info(f'{name} cloning {self._repo}')
-            print(f'git clone {self._repo_url}')
+        if self.repo not in self.run_command(name, ['ls'], print_output=False):
+            self._logger.info(f'{name} cloning {self.repo}')
             self.run_command(name, [f'git clone {self._repo_url}'])
 
         self._logger.info(f'{name} checking out {self._commit}')
         self.run_command(name, [
-            f'cd ./{self._repo}',
+            f'cd ./{self.repo}',
             'git fetch',
-            'git pull',
-            # f'git -c advice.detachedHead=false checkout {self._commit}'
+            'git pull'
         ])
 
     def _stop(self, name, data):
@@ -509,13 +563,12 @@ class EC2Manager:
         """
         Runs start command on a single instance.
         """
+        self._set_env_file(name, data)
         self._logger.info(f'{name} starting...')
         working_directory = self._get_working_directory(data['working_directory'])
         self.run_command(name, [
             f'cd {working_directory} ',
-            *[f'export {key}={value}' for key, value in data['envs'].items()],
             data['commands']['start'],
-            *[f'unset {key}' for key in data['envs'].keys()]
         ])
 
     def update_repos(self):
@@ -568,3 +621,10 @@ class EC2Manager:
         self.update_repos()
         self.stop()
         self.start()
+
+    def destroy(self):
+        """
+        Destroy resources created with the config.
+        """
+        self._init_terraform()
+        self._destroy_terraform()
